@@ -3,7 +3,7 @@
    Parser for dhcpd config file... */
 
 /*
- * Copyright (c) 2004-2006 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2008 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,12 +34,13 @@
 
 #ifndef lint
 static char copyright[] =
-"$Id: confpars.c,v 1.143.2.30 2006/02/22 22:43:27 dhankins Exp $ Copyright (c) 2004-2006 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: confpars.c,v 1.159.16.10 2008/01/22 19:02:51 dhankins Exp $ Copyright (c) 2004-2008 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
 
 static TIME parsed_time;
+static unsigned char global_host_once = 1;
 
 #if defined (TRACING)
 trace_type_t *trace_readconf_type;
@@ -266,6 +267,12 @@ isc_result_t lease_file_subparse (struct parse *cfile)
 			} else
 				parse_warn (cfile,
 					    "possibly corrupt lease file");
+		} else if (token == CLASS) {
+			parse_class_declaration(0, cfile, root_group,
+						CLASS_TYPE_CLASS);
+		} else if (token == SUBCLASS) {
+			parse_class_declaration(0, cfile, root_group,
+						CLASS_TYPE_SUBCLASS);
 		} else if (token == HOST) {
 			parse_host_declaration (cfile, root_group);
 		} else if (token == GROUP) {
@@ -288,8 +295,7 @@ isc_result_t lease_file_subparse (struct parse *cfile)
 
 /* statement :== parameter | declaration
 
-   parameter :== timestamp
-   	       | DEFAULT_LEASE_TIME lease_time
+   parameter :== DEFAULT_LEASE_TIME lease_time
 	       | MAX_LEASE_TIME lease_time
 	       | DYNAMIC_BOOTP_LEASE_CUTOFF date
 	       | DYNAMIC_BOOTP_LEASE_LENGTH lease_time
@@ -333,12 +339,13 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 	struct data_string data;
 	struct hardware hardware;
 	struct executable_statement *et, *ep;
-	struct option *option;
+	struct option *option = NULL;
 	struct option_cache *cache;
 	int lose;
 	struct data_string key_id;
 	int known;
 	isc_result_t status;
+	unsigned code;
 
 	token = peek_token (&val, (unsigned *)0, cfile);
 
@@ -359,9 +366,17 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 		
 	      case HOST:
 		next_token (&val, (unsigned *)0, cfile);
-		if (type != HOST_DECL && type != CLASS_DECL)
+		if (type != HOST_DECL && type != CLASS_DECL) {
+			if (global_host_once &&
+			    (type == SUBNET_DECL || type == SHARED_NET_DECL)) {
+				global_host_once = 0;
+				log_error("WARNING: Host declarations are "
+					  "global.  They are not limited to "
+					  "the scope you declared them in.");
+			}
+
 			parse_host_declaration (cfile, group);
-		else {
+		} else {
 			parse_warn (cfile,
 				    "host declarations not allowed here.");
 			skip_to_semi (cfile);
@@ -379,11 +394,6 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 		}
 		return 1;
 
-	      case TIMESTAMP:
-		next_token (&val, (unsigned *)0, cfile);
-		parsed_time = parse_timestamp (cfile);
-		break;
-
 	      case SHARED_NETWORK:
 		next_token (&val, (unsigned *)0, cfile);
 		if (type == SHARED_NET_DECL ||
@@ -459,7 +469,7 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 			skip_to_semi (cfile);
 			break;
 		}
-		parse_class_declaration ((struct class **)0, cfile, group, 0);
+		parse_class_declaration(NULL, cfile, group, CLASS_TYPE_VENDOR);
 		return 1;
 
 	      case USER_CLASS:
@@ -470,7 +480,7 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 			skip_to_semi (cfile);
 			break;
 		}
-		parse_class_declaration ((struct class **)0, cfile, group, 1);
+		parse_class_declaration(NULL, cfile, group, CLASS_TYPE_USER);
 		return 1;
 
 	      case CLASS:
@@ -481,7 +491,7 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 			skip_to_semi (cfile);
 			break;
 		}
-		parse_class_declaration ((struct class **)0, cfile, group, 2);
+		parse_class_declaration(NULL, cfile, group, CLASS_TYPE_CLASS);
 		return 1;
 
 	      case SUBCLASS:
@@ -492,12 +502,20 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 			skip_to_semi (cfile);
 			break;
 		}
-		parse_class_declaration ((struct class **)0, cfile, group, 3);
+		parse_class_declaration(NULL, cfile, group,
+					CLASS_TYPE_SUBCLASS);
 		return 1;
 
 	      case HARDWARE:
 		next_token (&val, (unsigned *)0, cfile);
 		memset (&hardware, 0, sizeof hardware);
+		if (host_decl && memcmp(&hardware, &(host_decl->interface),
+					sizeof(hardware)) != 0) {
+			parse_warn(cfile, "Host %s hardware address already "
+					  "configured.", host_decl->name);
+			break;
+		}
+
 		parse_hardware_param (cfile, &hardware);
 		if (host_decl)
 			host_decl -> interface = hardware;
@@ -532,11 +550,13 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 		next_token (&val, (unsigned *)0, cfile);
 		if (type != SUBNET_DECL && type != SHARED_NET_DECL) {
 			parse_warn (cfile, "pool declared outside of network");
-		}
-		if (type == POOL_DECL) {
+			skip_to_semi(cfile);
+		} else if (type == POOL_DECL) {
 			parse_warn (cfile, "pool declared within pool.");
-		}
-		parse_pool_statement (cfile, group, type);
+			skip_to_semi(cfile);
+		} else
+			parse_pool_statement (cfile, group, type);
+
 		return declaration;
 
 	      case RANGE:
@@ -576,7 +596,11 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 		/* "server-identifier" is a special hack, equivalent to
 		   "option dhcp-server-identifier". */
 	      case SERVER_IDENTIFIER:
-		option = dhcp_universe.options [DHO_DHCP_SERVER_IDENTIFIER];
+		code = DHO_DHCP_SERVER_IDENTIFIER;
+		if (!option_code_hash_lookup(&option, dhcp_universe.code_hash,
+					     &code, 0, MDL))
+			log_fatal("Server identifier not in hash (%s:%d).",
+				  MDL);
 		token = next_token (&val, (unsigned *)0, cfile);
 		goto finish_option;
 
@@ -596,8 +620,8 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 		}
 
 		known = 0;
-		option = parse_option_name (cfile, 1, &known);
-		if (option) {
+		status = parse_option_name(cfile, 1, &known, &option);
+		if (status == ISC_R_SUCCESS) {
 			token = peek_token (&val, (unsigned *)0, cfile);
 			if (token == CODE) {
 				if (type != ROOT_GROUP) {
@@ -605,13 +629,12 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 						    "option definitions%s",
 						    " may not be scoped.");
 					skip_to_semi (cfile);
-					free_option (option, MDL);
+					option_dereference(&option, MDL);
 					break;
 				}
 				next_token (&val, (unsigned *)0, cfile);
-				if (!parse_option_code_definition (cfile,
-								   option))
-					free_option (option, MDL);
+				parse_option_code_definition(cfile, option);
+				option_dereference(&option, MDL);
 				return declaration;
 			}
 
@@ -622,16 +645,36 @@ int parse_statement (cfile, group, type, host_decl, declaration)
 					    option -> universe -> name,
 					    option -> name);
 				skip_to_semi (cfile);
-				free_option (option, MDL);
+				option_dereference(&option, MDL);
 				return declaration;
 			}
 
+			/*
+			 * If the configuration attempts to define on option
+			 * that we ignore, then warn about it now.
+			 *
+			 * In DHCPv4 we do not use dhcp-renewal-time or
+			 * dhcp-rebinding-time, but we use these in DHCPv6.
+			 *
+			 * XXX: We may want to include a "blacklist" of 
+			 *      options we ignore in the future, as a table.
+			 */
+			if ((option->code == DHO_DHCP_LEASE_TIME) ||
+			    (option->code == DHO_DHCP_RENEWAL_TIME) ||
+			    (option->code == DHO_DHCP_REBINDING_TIME))
+			{
+				log_error("WARNING: server ignoring option %s "
+				          "in configuration file.",
+                                          option->name);
+			}
+
 		      finish_option:
 			et = (struct executable_statement *)0;
 			if (!parse_option_statement
 				(&et, cfile, 1, option,
 				 supersede_option_statement))
 				return declaration;
+			option_dereference(&option, MDL);
 			goto insert_statement;
 		} else
 			return declaration;
@@ -855,6 +898,22 @@ void parse_failover_peer (cfile, group, type)
 			cp -> port = atoi (val);
 			break;
 
+		      case MAX_LEASE_MISBALANCE:
+			tp = &peer->max_lease_misbalance;
+			goto parse_idle;
+
+		      case MAX_LEASE_OWNERSHIP:
+			tp = &peer->max_lease_ownership;
+			goto parse_idle;
+
+		      case MAX_BALANCE:
+			tp = &peer->max_balance;
+			goto parse_idle;
+
+		      case MIN_BALANCE:
+			tp = &peer->min_balance;
+			goto parse_idle;
+
 		      case MAX_RESPONSE_DELAY:
 			tp = &cp -> max_response_delay;
 		      parse_idle:
@@ -990,16 +1049,22 @@ void parse_failover_peer (cfile, group, type)
 			    "primary failover server must have mclt.");
 	    }
 	}
-	if (!peer -> me.max_flying_updates) {
-		peer -> me.max_flying_updates = 100;
-	}
-	if (!peer -> me.max_response_delay) {
-		peer -> me.max_response_delay = 60;
-	}
 
-	if (type == SHARED_NET_DECL) {
-		group -> shared_network -> failover_peer = peer;
-	}
+	if (!peer->max_lease_misbalance)
+		peer->max_lease_misbalance = DEFAULT_MAX_LEASE_MISBALANCE;
+	if (!peer->max_lease_ownership)
+		peer->max_lease_ownership = DEFAULT_MAX_LEASE_OWNERSHIP;
+	if (!peer->max_balance)
+		peer->max_balance = DEFAULT_MAX_BALANCE_TIME;
+	if (!peer->min_balance)
+		peer->min_balance = DEFAULT_MIN_BALANCE_TIME;
+	if (!peer->me.max_flying_updates)
+		peer->me.max_flying_updates = DEFAULT_MAX_FLYING_UPDATES;
+	if (!peer->me.max_response_delay)
+		peer->me.max_response_delay = DEFAULT_MAX_RESPONSE_DELAY;
+
+	if (type == SHARED_NET_DECL)
+		group->shared_network->failover_peer = peer;
 
 	/* Set the initial state. */
 	if (peer -> i_am == primary) {
@@ -1270,12 +1335,19 @@ void parse_pool_statement (cfile, group, type)
 		shared_network_reference (&pool -> shared_network,
 					  group -> subnet -> shared_network,
 					  MDL);
-	else
+	else if (type == SHARED_NET_DECL)
 		shared_network_reference (&pool -> shared_network,
 					  group -> shared_network, MDL);
+	else {
+		parse_warn(cfile, "Dynamic pools are only valid inside "
+				  "subnet or shared-network statements.");
+		skip_to_semi(cfile);
+		return;
+	}
 
-	if (!clone_group (&pool -> group, pool -> shared_network -> group, MDL))
-		log_fatal ("can't clone pool group.");
+	if (pool->shared_network == NULL ||
+            !clone_group(&pool->group, pool->shared_network->group, MDL))
+		log_fatal("can't clone pool group.");
 
 #if defined (FAILOVER_PROTOCOL)
 	/* Inherit the failover peer from the shared network. */
@@ -1433,7 +1505,7 @@ void parse_pool_statement (cfile, group, type)
 			      default:
 				parse_warn (cfile, "expecting permit type.");
 				skip_to_semi (cfile);
-				break;
+				goto cleanup;
 			}
 			while (*permit_head)
 				permit_head = &((*permit_head) -> next);
@@ -1450,6 +1522,15 @@ void parse_pool_statement (cfile, group, type)
 			done = 1;
 			break;
 
+		      case END_OF_FILE:
+			/*
+			 * We can get to END_OF_FILE if, for instance,
+			 * the parse_statement() reads all available tokens
+			 * and leaves us at the end.
+			 */
+			parse_warn(cfile, "unexpected end of file");
+			return;
+
 		      default:
 			declaration = parse_statement (cfile, pool -> group,
 						       POOL_DECL,
@@ -1459,60 +1540,6 @@ void parse_pool_statement (cfile, group, type)
 		}
 	} while (!done);
 
-#if defined (FAILOVER_PROTOCOL)
-	/* We can't do failover on a pool that supports dynamic bootp,
-	   because BOOTP doesn't support leases, and failover absolutely
-	   depends on lease timing. */
-	if (pool -> failover_peer) {
-		/* This search order matches the search orders later in
-		 * execution - deny first, if not denied, check permit
-		 * list.  A dynamic bootp client may be known or unknown,
-		 * it may belong to a member of a class, but it definitely
-		 * will not be authenticated since that requires DHCP
-		 * to work.  So a dynamic bootp client is definitely not
-		 * an authenticated client, and we can't say for sure about
-		 * anything else.
-		 *
-		 * So we nag the user.
-		 */
-		for (permit = pool -> prohibit_list; permit;
-		     permit = permit -> next) {
-			if (permit -> type == permit_dynamic_bootp_clients ||
-			    permit -> type == permit_unauthenticated_clients ||
-			    permit -> type == permit_all_clients)
-				break;
-		}
-		if (!permit) {
-			permit = pool -> permit_list;
-			do {
-				if (!permit ||
-				    permit -> type !=
-					permit_authenticated_clients) {
-					parse_warn (cfile,
-					  "pools with failover peers %s",
-					  "may not permit dynamic bootp.");
-					log_error ("Either write a \"%s\" %s",
-					  "no failover",
-					  "statement and use disjoint");
-					log_error ("pools, or%s (%s) %s",
-					  " don't permit dynamic bootp",
-					  "\"deny dynamic bootp clients;\"",
-					  "in this pool.");
-					log_error ("This is a protocol,%s %s",
-					   " limitation, not an ISC DHCP",
-					   "limitation, so");
-					log_error ("please don't request an %s",
-					   "enhancement or ask why this is.");
-
-					break;
-				}
-
-				permit = permit -> next;
-			} while (permit);
-		}
-	}
-#endif /* FAILOVER_PROTOCOL */
-
 	/* See if there's already a pool into which we can merge this one. */
 	for (pp = pool -> shared_network -> pools; pp; pp = pp -> next) {
 		struct lease *l;
@@ -1559,6 +1586,7 @@ void parse_pool_statement (cfile, group, type)
 		log_error ("one range statement.");
 	}
 
+cleanup:
 	/* Dereference the lease chain. */
 	lp = (struct lease *)0;
 	while (lpchain) {
@@ -1722,6 +1750,13 @@ void parse_host_declaration (cfile, group)
 			token = next_token (&val, (unsigned *)0, cfile);
 			data_string_forget (&host -> client_identifier, MDL);
 
+			if (host->client_identifier.len != 0) {
+				parse_warn(cfile, "Host %s already has a "
+						  "client identifier.",
+					   host->name);
+				break;
+			}
+
 			/* See if it's a string or a cshl. */
 			token = peek_token (&val, (unsigned *)0, cfile);
 			if (token == STRING) {
@@ -1823,6 +1858,9 @@ int parse_class_declaration (cp, cfile, group, type)
 	struct expression *expr;
 	int new = 1;
 	isc_result_t status = ISC_R_FAILURE;
+	int matchedonce = 0;
+	int submatchedonce = 0;
+	unsigned code;
 
 	token = next_token (&val, (unsigned *)0, cfile);
 	if (token != STRING) {
@@ -1834,18 +1872,17 @@ int parse_class_declaration (cp, cfile, group, type)
 	/* See if there's already a class with the specified name. */
 	find_class (&pc, val, MDL);
 
-	/* If this isn't a subclass, we're updating an existing class. */
-	if (pc && type != 0 && type != 1 && type != 3) {
-		class_reference (&class, pc, MDL);
+	/* If it is a class, we're updating it.  If it's any of the other
+	 * types (subclass, vendor or user class), the named class is a
+	 * reference to the parent class so its mandatory.
+	 */
+	if (pc && (type == CLASS_TYPE_CLASS)) {
+		class_reference(&class, pc, MDL);
 		new = 0;
-		class_dereference (&pc, MDL);
-	}
-
-	/* If this _is_ a subclass, there _must_ be a class with the
-	   same name. */
-	if (!pc && (type == 0 || type == 1 || type == 3)) {
-		parse_warn (cfile, "no class named %s", val);
-		skip_to_semi (cfile);
+		class_dereference(&pc, MDL);
+	} else if (!pc && (type != CLASS_TYPE_CLASS)) {
+		parse_warn(cfile, "no class named %s", val);
+		skip_to_semi(cfile);
 		return 0;
 	}
 
@@ -1856,7 +1893,7 @@ int parse_class_declaration (cp, cfile, group, type)
 	   are turned into subclasses of the implicit classes, and the
 	   submatch expression of the implicit classes extracts the contents of
 	   the vendor class or user class. */
-	if (type == 0 || type == 1) {
+	if ((type == CLASS_TYPE_VENDOR) || (type == CLASS_TYPE_USER)) {
 		data.len = strlen (val);
 		data.buffer = (struct buffer *)0;
 		if (!buffer_allocate (&data.buffer, data.len + 1, MDL))
@@ -1865,7 +1902,7 @@ int parse_class_declaration (cp, cfile, group, type)
 		data.terminated = 1;
 
 		tname = type ? "implicit-vendor-class" : "implicit-user-class";
-	} else if (type == 2) {
+	} else if (type == CLASS_TYPE_CLASS) {
 		tname = val;
 	} else {
 		tname = (const char *)0;
@@ -1880,7 +1917,7 @@ int parse_class_declaration (cp, cfile, group, type)
 		name = (char *)0;
 
 	/* If this is a straight subclass, parse the hash string. */
-	if (type == 3) {
+	if (type == CLASS_TYPE_SUBCLASS) {
 		token = peek_token (&val, (unsigned *)0, cfile);
 		if (token == STRING) {
 			token = next_token (&val, &data.len, cfile);
@@ -1899,7 +1936,8 @@ int parse_class_declaration (cp, cfile, group, type)
 		} else if (token == NUMBER_OR_NAME || token == NUMBER) {
 			memset (&data, 0, sizeof data);
 			if (!parse_cshl (&data, cfile)) {
-				class_dereference (&pc, MDL);
+				if (pc)
+					class_dereference (&pc, MDL);
 				return 0;
 			}
 		} else {
@@ -1912,7 +1950,7 @@ int parse_class_declaration (cp, cfile, group, type)
 
 	/* See if there's already a class in the hash table matching the
 	   hash data. */
-	if (type == 0 || type == 1 || type == 3)
+	if (type != CLASS_TYPE_CLASS)
 		class_hash_lookup (&class, pc -> hash,
 				   (const char *)data.data, data.len, MDL);
 
@@ -1936,13 +1974,15 @@ int parse_class_declaration (cp, cfile, group, type)
 			}
 			data_string_copy (&class -> hash_string, &data, MDL);
 			if (!pc -> hash &&
-			    !class_new_hash (&pc -> hash, 0, MDL))
+			    !class_new_hash (&pc->hash, SCLASS_HASH_SIZE, MDL))
 				log_fatal ("No memory for subclass hash.");
 			class_hash_add (pc -> hash,
 					(const char *)class -> hash_string.data,
 					class -> hash_string.len,
 					(void *)class, MDL);
 		} else {
+			if (class->group)
+				group_dereference(&class->group, MDL);
 			if (!clone_group (&class -> group, group, MDL))
 				log_fatal ("no memory to clone class group.");
 		}
@@ -1950,7 +1990,7 @@ int parse_class_declaration (cp, cfile, group, type)
 		/* If this is an implicit vendor or user class, add a
 		   statement that causes the vendor or user class ID to
 		   be sent back in the reply. */
-		if (type == 0 || type == 1) {
+		if (type == CLASS_TYPE_VENDOR || type == CLASS_TYPE_USER) {
 			stmt = (struct executable_statement *)0;
 			if (!executable_statement_allocate (&stmt, MDL))
 				log_fatal ("no memory for class statement.");
@@ -1958,21 +1998,25 @@ int parse_class_declaration (cp, cfile, group, type)
 			if (option_cache_allocate (&stmt -> data.option,
 						   MDL)) {
 				stmt -> data.option -> data = data;
-				stmt -> data.option -> option =
-					dhcp_universe.options
-					[type
+				code = (type == CLASS_TYPE_VENDOR)
 					? DHO_VENDOR_CLASS_IDENTIFIER
-					: DHO_USER_CLASS];
+					: DHO_USER_CLASS;
+				option_code_hash_lookup(
+						&stmt->data.option->option,
+							dhcp_universe.code_hash,
+							&code, 0, MDL);
 			}
 			class -> statements = stmt;
 		}
 
 		/* Save the name, if there is one. */
-		class -> name = name;
+		if (class->name != NULL)
+			dfree(class->name, MDL);
+		class->name = name;
 	}
 
-	if (type == 0 || type == 1 || type == 3)
-		data_string_forget (&data, MDL);
+	if (type != CLASS_TYPE_CLASS)
+		data_string_forget(&data, MDL);
 
 	/* Spawned classes don't have to have their own settings. */
 	if (class -> superclass) {
@@ -2008,6 +2052,18 @@ int parse_class_declaration (cp, cfile, group, type)
 			token = next_token (&val, (unsigned *)0, cfile);
 			parse_warn (cfile, "unexpected end of file");
 			break;
+		} else if (token == DYNAMIC) {
+			class->flags |= CLASS_DECL_DYNAMIC;
+			token = next_token (&val, (unsigned *)0, cfile);
+			if (!parse_semi (cfile))
+				break;
+			continue;
+		} else if (token == TOKEN_DELETED) {
+			class->flags |= CLASS_DECL_DELETED;
+			token = next_token (&val, (unsigned *)0, cfile);
+			if (!parse_semi (cfile))
+				break;
+			continue;
 		} else if (token == MATCH) {
 			if (pc) {
 				parse_warn (cfile,
@@ -2019,13 +2075,17 @@ int parse_class_declaration (cp, cfile, group, type)
 			token = peek_token (&val, (unsigned *)0, cfile);
 			if (token != IF)
 				goto submatch;
-			if (class -> expr) {
-				parse_warn (cfile, "can't override match.");
-				skip_to_semi (cfile);
+			token = next_token (&val, (unsigned *)0, cfile);
+			if (matchedonce) {
+				parse_warn(cfile, "A class may only have "
+						  "one 'match if' clause.");
+				skip_to_semi(cfile);
 				break;
 			}
-			token = next_token (&val, (unsigned *)0, cfile);
-			if (!parse_boolean_expression (&class -> expr, cfile,
+			matchedonce = 1;
+			if (class->expr)
+				expression_dereference(&class->expr, MDL);
+			if (!parse_boolean_expression (&class->expr, cfile,
 						       &lose)) {
 				if (!lose) {
 					parse_warn (cfile,
@@ -2040,13 +2100,13 @@ int parse_class_declaration (cp, cfile, group, type)
 				parse_semi (cfile);
 			}
 		} else if (token == SPAWN) {
+			token = next_token (&val, (unsigned *)0, cfile);
 			if (pc) {
 				parse_warn (cfile,
 					    "invalid spawn in subclass.");
 				skip_to_semi (cfile);
 				break;
 			}
-			token = next_token (&val, (unsigned *)0, cfile);
 			class -> spawning = 1;
 			token = next_token (&val, (unsigned *)0, cfile);
 			if (token != WITH) {
@@ -2056,13 +2116,16 @@ int parse_class_declaration (cp, cfile, group, type)
 				break;
 			}
 		      submatch:
-			if (class -> submatch) {
+			if (submatchedonce) {
 				parse_warn (cfile,
 					    "can't override existing %s.",
 					    "submatch/spawn");
 				skip_to_semi (cfile);
 				break;
 			}
+			submatchedonce = 1;
+			if (class->submatch)
+				expression_dereference(&class->submatch, MDL);
 			if (!parse_data_expression (&class -> submatch,
 						    cfile, &lose)) {
 				if (!lose) {
@@ -2094,6 +2157,8 @@ int parse_class_declaration (cp, cfile, group, type)
 				break;
 			}
 			class -> lease_limit = atoi (val);
+			if (class->billed_leases)
+				dfree(class->billed_leases, MDL);
 			class -> billed_leases =
 				dmalloc (class -> lease_limit *
 					 sizeof (struct lease *), MDL);
@@ -2111,7 +2176,22 @@ int parse_class_declaration (cp, cfile, group, type)
 						       declaration);
 		}
 	} while (1);
-	if (type == 2 && new) {
+
+	if (class->flags & CLASS_DECL_DELETED) {
+		if (type == CLASS_TYPE_CLASS) {
+			struct class *theclass = NULL;
+		
+			status = find_class(&theclass, class->name, MDL);
+			if (status == ISC_R_SUCCESS) {
+				delete_class(theclass, 0);
+				class_dereference(&theclass, MDL);
+			}
+		} else {
+			class_hash_delete(pc->hash,
+					  (char *)class->hash_string.data,
+					  class->hash_string.len, MDL);
+		}
+	} else if (type == CLASS_TYPE_CLASS && new) {
 		if (!collections -> classes)
 			class_reference (&collections -> classes, class, MDL);
 		else {
@@ -2122,7 +2202,8 @@ int parse_class_declaration (cp, cfile, group, type)
 			class_reference (&c -> nic, class, MDL);
 		}
 	}
-	if (cp)
+
+	if (cp)				/* should always be 0??? */
 		status = class_reference (cp, class, MDL);
 	class_dereference (&class, MDL);
 	if (pc)
@@ -2481,21 +2562,6 @@ int parse_fixed_addr_param (oc, cfile)
 	return status;
 }
 
-/* timestamp :== date
-
-   Timestamps are actually not used in dhcpd.conf, which is a static file,
-   but rather in the database file and the journal file.  (Okay, actually
-   they're not even used there yet). */
-
-TIME parse_timestamp (cfile)
-	struct parse *cfile;
-{
-	TIME rv;
-
-	rv = parse_date (cfile);
-	return rv;
-}
-		
 /* lease_declaration :== LEASE ip_address LBRACE lease_parameters RBRACE
 
    lease_parameters :== <nil>
@@ -2587,11 +2653,6 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 				lease -> ends = t;
 				break;
 				
-			      case TIMESTAMP:
-				seenbit = 4;
-				lease -> timestamp = t;
-				break;
-
 			      case TSTP:
 				seenbit = 65536;
 				lease -> tstp = t;
@@ -2618,7 +2679,7 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 			}
 			break;
 
-			/* Colon-seperated hexadecimal octets... */
+			/* Colon-separated hexadecimal octets... */
 		      case UID:
 			seenbit = 8;
 			token = peek_token (&val, (unsigned *)0, cfile);
@@ -2667,7 +2728,7 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 				log_fatal ("No memory for lease uid");
 			}
 			break;
-			
+
 		      case CLASS:
 			seenbit = 32;
 			token = next_token (&val, (unsigned *)0, cfile);
@@ -2687,12 +2748,19 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 					      &lease -> hardware_addr);
 			break;
 
+		      case TOKEN_RESERVED:
+			seenbit = 0;
+			lease->flags |= RESERVED_LEASE;
+			parse_semi(cfile);
+			break;
+
 		      case DYNAMIC_BOOTP:
-			seenbit = 256;
+			seenbit = 0;
 			lease -> flags |= BOOTP_LEASE;
 			parse_semi (cfile);
 			break;
-			
+
+			/* XXX: Reverse compatibility? */
 		      case TOKEN_ABANDONED:
 			seenbit = 256;
 			lease -> binding_state = FTS_ABANDONED;
@@ -2743,13 +2811,19 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 			      case TOKEN_BACKUP:
 				new_state = FTS_BACKUP;
 				break;
+
+				/* RESERVED and BOOTP states preserved for
+				 * compatiblity with older versions.
+				 */
 			      case TOKEN_RESERVED:
 				new_state = FTS_ACTIVE;
+				lease->flags |= RESERVED_LEASE;
 				break;
 			      case TOKEN_BOOTP:
 				new_state = FTS_ACTIVE;
-				lease -> flags |= BOOTP_LEASE;
+				lease->flags |= BOOTP_LEASE;
 				break;
+
 			      default:
 				parse_warn (cfile,
 					    "%s: expecting a binding state.",
@@ -2822,9 +2896,8 @@ int parse_lease_declaration (struct lease **lp, struct parse *cfile)
 				if (lease -> billing_class)
 				    class_dereference (&lease -> billing_class,
 						       MDL);
-				parse_class_declaration
-					(&class,
-					 cfile, (struct group *)0, 3);
+				parse_class_declaration(&class, cfile, NULL,
+							CLASS_TYPE_SUBCLASS);
 			} else {
 				parse_warn (cfile, "expecting \"class\"");
 				if (token != SEMI)
@@ -3268,6 +3341,8 @@ int parse_allow_deny (oc, cfile, flag)
 	enum dhcp_token token;
 	const char *val;
 	unsigned char rf = flag;
+	unsigned code;
+	struct option *option = NULL;
 	struct expression *data = (struct expression *)0;
 	int status;
 
@@ -3277,42 +3352,35 @@ int parse_allow_deny (oc, cfile, flag)
 	token = next_token (&val, (unsigned *)0, cfile);
 	switch (token) {
 	      case TOKEN_BOOTP:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_ALLOW_BOOTP], MDL);
+		code = SV_ALLOW_BOOTP;
 		break;
 
 	      case BOOTING:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_ALLOW_BOOTING],
-				       MDL);
+		code = SV_ALLOW_BOOTING;
 		break;
 
 	      case DYNAMIC_BOOTP:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_DYNAMIC_BOOTP],
-				       MDL);
+		code = SV_DYNAMIC_BOOTP;
 		break;
 
 	      case UNKNOWN_CLIENTS:
-		status = (option_cache
-			  (oc, (struct data_string *)0, data,
-			   &server_options [SV_BOOT_UNKNOWN_CLIENTS], MDL));
+		code = SV_BOOT_UNKNOWN_CLIENTS;
 		break;
 
 	      case DUPLICATES:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_DUPLICATES], MDL);
+		code = SV_DUPLICATES;
 		break;
 
 	      case DECLINES:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_DECLINES], MDL);
+		code= SV_DECLINES;
 		break;
 
 	      case CLIENT_UPDATES:
-		status = option_cache (oc, (struct data_string *)0, data,
-				       &server_options [SV_CLIENT_UPDATES],
-				       MDL);
+		code = SV_CLIENT_UPDATES;
+		break;
+
+	      case LEASEQUERY:
+		code = SV_LEASEQUERY;
 		break;
 
 	      default:
@@ -3320,6 +3388,12 @@ int parse_allow_deny (oc, cfile, flag)
 		skip_to_semi (cfile);
 		return 0;
 	}
+	/* Reference on option is passed to option cache. */
+	if (!option_code_hash_lookup(&option, server_universe.code_hash,
+				     &code, 0, MDL))
+		log_fatal("Unable to find server option %u (%s:%d).",
+			  code, MDL);
+	status = option_cache(oc, NULL, data, option, MDL);
 	expression_dereference (&data, MDL);
 	parse_semi (cfile);
 	return status;
